from tflite_runtime.interpreter import Interpreter
import os
import sys
import cv2
import numpy as np
import argparse
import time

DIR_NAME = os.path.dirname(__file__)

MODEL_NAME = 'human_segmentation.tflite'

DEMO_WINDOW_NAME = 'Image Segmentation Demo'

# layout settings for armadillo-demo-experience
CAMERA_WIDTH = 1100
CAMERA_HEIGHT = 600
OVERLAY_CAMERA_XPOS = 0
OVERLAY_CAMERA_YPOS = 0

class Inferencer:
    def __init__(self, model_path, camera):
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, self.input_height, self.input_width, _ = self.input_details[0]['shape']
        self.input_dtype = self.input_details[0]['dtype']
        self.camera = camera


    def _preprocess_input(self, image):
        resized = cv2.resize(image, dsize=(self.input_width, self.input_height))
        image_for_prediction = np.asarray(resized / 127.5 -1).astype(self.input_dtype)
        return np.expand_dims(image_for_prediction, 0)


    def _overlay_imgs(self, bg, fg, xpos, ypos):
        frame = bg
        height, width = fg.shape[:2]
        frame[ypos:height + ypos, xpos:width + xpos] = fg
        return frame


    def infer(
            self,
            target_window,
            cw,
            ch,
            ocx=OVERLAY_CAMERA_XPOS,
            ocy=OVERLAY_CAMERA_YPOS,
            back_ground_image=None
    ):
        if back_ground_image is None:
            back_ground_image = np.zeros((ch, cw, 3), np.uint8)
            ocx = 0
            ocy = 0

        while True:
            success, image = self.camera.read()
            if not success:
                sys.exit(
                    'ERROR: Unable to read from webcam. Please verify your webcam settings.'
                )

            input_img = self._preprocess_input(image)

            start = time.monotonic()
            self.interpreter.set_tensor(self.input_details[0]['index'], input_img)
            self.interpreter.invoke()
            elapsed_time = time.monotonic() - start

            org_height, org_width, _ = image.shape

            res = self.interpreter.get_tensor(self.output_details[0]['index'])
            output = np.squeeze(res)
            blue = (output == 1)
            output = np.zeros(list(output.shape) + [3])
            output[blue] = [255, 0, 0]
            output = cv2.resize(output, dsize=(org_width, org_height)).astype(np.uint8)

            frame = cv2.addWeighted(image, 1, output, 0.8, 0)
            cv2.putText(
                frame,
                text='FPS(inference only): {:.2f}'.format(1 / elapsed_time),
                org=(10, 15),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.5,
                color=(0, 0, 255),
                thickness=1,
                lineType=cv2.LINE_AA
            )

            frame = self._overlay_imgs(
                back_ground_image,
                cv2.resize(frame, dsize=(cw, ch)),
                ocx,
                ocy
            )
            cv2.imshow(target_window, frame)
            cv2.waitKey(1)


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--model',
        help='File path of .tflite file.',
        required=False,
        default=MODEL_NAME
    )
    parser.add_argument(
        '--camera_id',
        type=int,
        default=2
    )
    parser.add_argument(
        '--width',
        type=int,
        default=CAMERA_WIDTH
    )
    parser.add_argument(
        '--height',
        type=int,
        default=CAMERA_HEIGHT
    )
    parser.add_argument(
        '--file',
        type=str,
        default=None,
        help="Optionally use a video file instead of a live camera"
    )
    args = parser.parse_args()

    model = os.path.join(DIR_NAME, args.model)

    if args.file is not None:
        cap = cv2.VideoCapture(args.file)
    else:
        cap = cv2.VideoCapture(args.camera_id + cv2.CAP_V4L2)
    cap.set(cv2.CAP_PROP_FRAME_WIDTH, args.width)
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, args.height)

    inferencer = Inferencer(model, cap)

    try:
        inferencer.infer(
            target_window=DEMO_WINDOW_NAME,
            cw=args.width,
            ch=args.height
        )
    except KeyboardInterrupt:
        cap.release()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
