# python3
#
# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Example using TF Lite to detect objects with the Raspberry Pi camera."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse
import io
import re
import time
import os
import sys

import cv2
import numpy as np

from PIL import Image
from tflite_runtime.interpreter import Interpreter

sys.path.append(os.path.dirname(__file__))
from annotation import Annotator

DIR_NAME = os.path.dirname(__file__)

MODEL_NAME = 'detect.tflite'
LABEL_NAME = 'coco_labels.txt'

DEMO_BACKGROUND = os.path.join(DIR_NAME, 'bg.jpg')
DEMO_WINDOW_NAME = 'Object Detection Demo'

# layout settings for armadillo-demo-experience
CAMERA_WIDTH = 800
CAMERA_HEIGHT = 600
OBJLIST_WIDTH = 300
OBJLIST_HEIGHT = 600

OVERLAY_CAMERA_XPOS = 0
OVERLAY_CAMERA_YPOS = 0
OVERLAY_OBJLIST_XPOS = OVERLAY_CAMERA_XPOS + CAMERA_WIDTH
OVERLAY_OBJLIST_YPOS = 0


class Inferencer:
    def __init__(
            self,
            model_path,
            label_path,
            threshold,
            camera
    ):
        self.labels = self._load_labels(label_path)
        self.threshold = threshold
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        _, self.input_height, self.input_width, _ = self.interpreter.get_input_details()[0]['shape']
        self.camera = camera
        self.annotator = Annotator(self.camera, default_color=(0, 0, 128))

    def _load_labels(self, path):
        """Loads the labels file. Supports files with or without index numbers."""
        labels = {}
        with open(path, 'r', encoding='utf-8') as f:
            lines = f.readlines()
            for row_number, content in enumerate(lines):
                pair = re.split(r'[:\s]+', content.strip(), maxsplit=1)
                if len(pair) == 2 and pair[0].strip().isdigit():
                    labels[int(pair[0])] = pair[1].strip()
                else:
                    labels[row_number] = pair[0].strip()
        return labels

    def infer(
            self,
            back_ground_image,
            target_window=DEMO_WINDOW_NAME,
            cw=CAMERA_WIDTH,
            ch=CAMERA_HEIGHT,
            ow=OBJLIST_WIDTH,
            oh=OBJLIST_HEIGHT,
            ocx=OVERLAY_CAMERA_XPOS,
            ocy=OVERLAY_CAMERA_YPOS,
            oox=OVERLAY_OBJLIST_XPOS,
            ooy=OVERLAY_OBJLIST_YPOS
    ):
        canvas = ObjList(oh, ow)
        while True:
            self.annotator.update()
            image = self.annotator.buffer.resize((self.input_width, self.input_height), Image.BICUBIC)
            start_time = time.monotonic()
            results = self._detect_objects(self.interpreter, image, self.threshold)
            elapsed_ms = (time.monotonic() - start_time) * 1000
            canvas.print_results(results, self.labels)
            self._annotate_objects(results, self.labels, cw, ch)
            self.annotator.text([5, 0], '%.1fms' % (elapsed_ms))
            frame = cv2.cvtColor(np.asarray(self.annotator.buffer), cv2.COLOR_RGB2BGR)
            frame = self._overlay_imgs(
                back_ground_image,
                cv2.resize(frame, dsize=(cw, ch)),
                ocx,
                ocy
            )
            frame = self._overlay_imgs(frame, canvas.canvas, oox, ooy)
            cv2.imshow(target_window, frame)
            cv2.waitKey(1)
            del image

    def _detect_objects(self, interpreter, image, threshold):
        """Returns a list of detection results, each a dictionary of object info."""
        self._set_input_tensor(interpreter, image)
        interpreter.invoke()

        # Get all output details
        boxes = self._get_output_tensor(interpreter, 0)
        classes = self._get_output_tensor(interpreter, 1)
        scores = self._get_output_tensor(interpreter, 2)
        count = int(self._get_output_tensor(interpreter, 3))

        results = []
        for i in range(count):
            if scores[i] >= threshold:
                result = {
                    'bounding_box': boxes[i],
                    'class_id': classes[i],
                    'score': scores[i]
                }
                results.append(result)
        return results

    def _set_input_tensor(self, interpreter, image):
        """Sets the input tensor."""
        tensor_index = interpreter.get_input_details()[0]['index']
        input_tensor = interpreter.tensor(tensor_index)()[0]
        input_tensor[:, :] = image


    def _get_output_tensor(self, interpreter, index):
        """Returns the output tensor at the given index."""
        output_details = interpreter.get_output_details()[index]
        tensor = np.squeeze(interpreter.get_tensor(output_details['index']))
        return tensor

    def _annotate_objects(self, results, labels, camera_width, camera_height):
        """Draws the bounding box and label for each object in the results."""
        for obj in results:
            # Convert the bounding box figures from relative coordinates
            # to absolute coordinates based on the original resolution
            ymin, xmin, ymax, xmax = obj['bounding_box']
            xmin = int(xmin * camera_width)
            xmax = int(xmax * camera_width)
            ymin = int(ymin * camera_height)
            ymax = int(ymax * camera_height)

            # Overlay the box, label, and score on the camera preview
            self.annotator.bounding_box([xmin, ymin, xmax, ymax])
            self.annotator.text([xmin, ymin],
                           '%s\n%.2f' % (labels[obj['class_id']], obj['score']))

    def _overlay_imgs(self, bg, fg, xpos, ypos):
        frame = bg
        height, width = fg.shape[:2]
        frame[ypos:height + ypos, xpos:width + xpos] = fg
        return frame


class ObjList:
    def __init__(self, height, width):
        self.height = height
        self.width = width
        self.clear_canvas()


    def clear_canvas(self):
        self.canvas = np.zeros((self.height, self.width, 3), dtype='uint8')
        self.canvas[:] = [128, 128, 128]


    def print_results(self, results, labels):
        ypos = 30
        self.clear_canvas()
        for obj in results:
            cv2.putText(self.canvas,
                        '{:15s}'.format(labels[obj['class_id']]),
                        (10, ypos),
                        cv2.FONT_HERSHEY_COMPLEX,
                        1,
                        (0, 0, 0),
                        2,
                        cv2.LINE_AA
            )
            cv2.putText(self.canvas,
                        '{:.2f}'.format(obj['score']),
                        (220, ypos),
                        cv2.FONT_HERSHEY_COMPLEX,
                        1,
                        (0, 0, 0),
                        2,
                        cv2.LINE_AA
            )
            ypos += 40


def calc_width(width):
    camera_width = int(width * (3 / 4))
    list_width = int(width * (1 / 4))
    sub = width - (camera_width + list_width)
    if sub != 0:
        list_width += sub
    return (camera_width, list_width)


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--model',
        help='File path of .tflite file.',
        required=False,
        default=os.path.join(DIR_NAME, MODEL_NAME))
    parser.add_argument(
        '--labels',
        help='File path of labels file.',
        required=False,
        default=os.path.join(DIR_NAME, LABEL_NAME))
    parser.add_argument(
        '--threshold',
        help='Score threshold for detected objects.',
        required=False,
        type=float,
        default=0.4)
    parser.add_argument(
        '--camera_id',
        help='video device number N (ex. /dev/videoN)',
        required=False,
        type=int,
        default=5)
    parser.add_argument(
        '--width',
        help='width of the output display (px)',
        required=False,
        type=int,
        default=CAMERA_WIDTH + OBJLIST_WIDTH)
    parser.add_argument(
        '--height',
        help='height of the output display (px)',
        required=False,
        type=int,
        default=CAMERA_HEIGHT)
    args = parser.parse_args()

    camera = cv2.VideoCapture(args.camera_id + cv2.CAP_V4L2)

    is_specified_size = False
    camera_width = CAMERA_WIDTH
    camera_height = CAMERA_HEIGHT
    list_width = OBJLIST_WIDTH
    if args.width != (CAMERA_WIDTH + OBJLIST_WIDTH):
        is_specified_size = True
        camera_width, list_width = calc_width(args.width)
    if args.height != CAMERA_HEIGHT:
        is_specified_size = True
        camera_height = args.height

    if camera.isOpened() is False:
        raise("IO Error")
    try:
        camera.set(cv2.CAP_PROP_FRAME_WIDTH, camera_width)
        camera.set(cv2.CAP_PROP_FRAME_HEIGHT, args.height)

        inferencer = Inferencer(args.model, args.labels, args.threshold, camera)

        cv2.namedWindow(DEMO_WINDOW_NAME, cv2.WINDOW_NORMAL)
        bg = cv2.imread(DEMO_BACKGROUND)
        if is_specified_size:
            bg = cv2.resize(bg, dsize=(camera_width + list_width, camera_height))
        cv2.imshow(DEMO_WINDOW_NAME, bg)

        if is_specified_size:
            inferencer.infer(
                bg,
                DEMO_WINDOW_NAME,
                camera_width,
                camera_height,
                list_width,
                camera_height,
                OVERLAY_CAMERA_XPOS,
                OVERLAY_CAMERA_YPOS,
                OVERLAY_CAMERA_XPOS + camera_width,
                OVERLAY_OBJLIST_YPOS
            )
        else:
            inferencer.infer(bg)
    except KeyboardInterrupt:
        camera.release()
        cv2.destroyAllWindows()


if __name__ == '__main__':
    main()
