from tflite_runtime.interpreter import Interpreter
import os
import sys
import cv2
import numpy as np
import argparse
import time

sys.path.append(os.path.dirname(__file__))
import posenet

DIR_NAME = os.path.dirname(__file__)

MODEL_NAME = 'posenet.tflite'

DEMO_WINDOW_NAME = 'Pose Estimation Demo'

INDEX_HEATMAP = 3
INDEX_OFFSET = 2
INDEX_DISPLACEMENT_FWD = 1
INDEX_DISPLACEMENT_BWD = 0

OUTPUT_STRIDE = 16

# layout settings for armadillo-demo-experience
CAMERA_WIDTH = 1100
CAMERA_HEIGHT = 600
OVERLAY_CAMERA_XPOS = 0
OVERLAY_CAMERA_YPOS = 0

class Inferencer:
    def __init__(self, model_path, camera, scale_factor):
        self.interpreter = Interpreter(model_path)
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, self.input_height, self.input_width, _ = self.input_details[0]['shape']
        self.camera = camera
        self.scale_factor = scale_factor


    def _get_outputs(self, index):
        return self.interpreter.get_tensor(self.output_details[index]['index'])


    def _process_input(self, source_img, output_stride=16):
        target_width, target_height = posenet.valid_resolution(
            source_img.shape[1] * self.scale_factor,
            source_img.shape[0] * self.scale_factor,
            output_stride=output_stride
        )

        scale = np.array(
            [source_img.shape[0] / target_height,
             source_img.shape[1] / target_width]
        )

        input_img = cv2.resize(
            source_img,
            (target_width, target_height),
            interpolation=cv2.INTER_LINEAR
        )
        input_img = cv2.cvtColor(input_img, cv2.COLOR_BGR2RGB).astype(np.float32)
        return input_img, source_img, scale


    def _overlay_imgs(self, bg, fg, xpos, ypos):
        frame = bg
        height, width = fg.shape[:2]
        frame[ypos:height + ypos, xpos:width + xpos] = fg
        return frame


    def infer(
            self,
            target_window,
            cw,
            ch,
            ocx=OVERLAY_CAMERA_XPOS,
            ocy=OVERLAY_CAMERA_YPOS,
            back_ground_image=None
    ):
        if back_ground_image is None:
            back_ground_image = np.zeros((ch, cw, 3), np.uint8)
            ocx = 0
            ocy = 0

        while True:
            success, image = self.camera.read()
            if not success:
                sys.exit(
                    'ERROR: Unable to read from webcam. Please verify your webcam settings.'
                )

            input_image, display_image, output_scale = self._process_input(
                image,
                output_stride=OUTPUT_STRIDE
            )

            infer_scale = \
                np.array(
                    [input_image.shape[0] / self.input_height,
                     input_image.shape[1] / self.input_width]
                )
            resized = cv2.resize(
                input_image,
                dsize=(self.input_height, self.input_width)
            )
            img = np.expand_dims(resized, axis=0).astype(np.float32)

            start = time.monotonic()
            self.interpreter.set_tensor(self.input_details[0]['index'], img)
            self.interpreter.invoke()
            heatmaps_result = self._get_outputs(INDEX_HEATMAP)
            offsets_result = self._get_outputs(INDEX_OFFSET)
            displacement_fwd_result = self._get_outputs(INDEX_DISPLACEMENT_FWD)
            displacement_bwd_result = self._get_outputs(INDEX_DISPLACEMENT_BWD)
            elapsed_time = time.monotonic() - start

            pose_scores, keypoint_scores, keypoint_coords = \
                posenet.decode_multi.decode_multiple_poses(
                    heatmaps_result.squeeze(axis=0),
                    offsets_result.squeeze(axis=0),
                    displacement_fwd_result.squeeze(axis=0),
                    displacement_bwd_result.squeeze(axis=0),
                    output_stride=16,
                    max_pose_detections=10,
                    min_pose_score=0.15
                )

            keypoint_coords *= output_scale * infer_scale

            frame = posenet.draw_skel_and_kp(
                display_image,
                pose_scores,
                keypoint_scores,
                keypoint_coords,
                min_pose_score=0.30,
                min_part_score=0.1
            )

            cv2.putText(
                frame,
                text='FPS(inference only): {:.2f}'.format(1 / elapsed_time),
                org=(10, 15),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.5,
                color=(0, 0, 255),
                thickness=1,
                lineType=cv2.LINE_AA
            )

            frame = self._overlay_imgs(
                back_ground_image,
                cv2.resize(frame, dsize=(cw, ch)),
                ocx,
                ocy
            )
            cv2.imshow(target_window, frame)
            cv2.waitKey(1)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--model',
        type=str,
        default=MODEL_NAME
    )
    parser.add_argument(
        '--camera_id',
        type=int,
        default=2
    )
    parser.add_argument(
        '--width',
        type=int,
        default=CAMERA_WIDTH
    )
    parser.add_argument(
        '--height',
        type=int,
        default=CAMERA_HEIGHT
    )
    parser.add_argument(
        '--scale_factor',
        type=float,
        default=0.7125
    )
    parser.add_argument(
        '--file',
        type=str,
        default=None,
        help="Optionally use a video file instead of a live camera"
    )
    args = parser.parse_args()

    model = os.path.join(DIR_NAME, args.model)

    if args.file is not None:
        cap = cv2.VideoCapture(args.file)
    else:
        cap = cv2.VideoCapture(args.camera_id + cv2.CAP_V4L2)
    cap.set(cv2.CAP_PROP_FRAME_WIDTH, args.width)
    cap.set(cv2.CAP_PROP_FRAME_HEIGHT, args.height)

    inferencer = Inferencer(model, cap, args.scale_factor)

    try:
        inferencer.infer(
            target_window=DEMO_WINDOW_NAME,
            cw=args.width,
            ch=args.height
        )
    except KeyboardInterrupt:
        cap.release()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    main()
